/*
 * USB [C]ommunications [D]evice [C]lass
 * RGB LED Cube Controller
 *
 *
 *
 * (c)2010
 * Oliver Kster
 * 1772570
 * Hochschule fr angewandte Wissenschaften Hamburg
 *
 */

#include "LED_Cube_AT90USB.h"
#include "Descriptors.h"

#define R 0
#define G 1
#define B 2


/*
 ***************************************************************************
 *
 * Globale Variablen / Speicherbereiche
 *
 ***************************************************************************
 */
static FILE USBSerialStream;

uint8_t input_stream[4] = {0};

//static volatile uint16_t framebuffer[8][64] = {{0}};
static volatile uint8_t framebuffer[8][64][3] = {{{0}}}; //Zugriff auf 8-Bit Variable ist schneller

static volatile uint8_t input_row = 0;
static volatile uint8_t set_row = 0;
static volatile uint8_t input_led = 0;
static volatile uint8_t control_byte = 0;

static volatile uint8_t setbit = 0;
static volatile uint8_t clearbit = 0;

static volatile uint8_t USBIsConnected = 0;

static volatile uint8_t animation_enabled = 0;
static volatile uint8_t animation_speed = 0;

int row = 1;

/*
 ***************************************************************************
 *
 * CDC Klassen-Treiber Interface-Konfiguration
 *
 ***************************************************************************
 */
USB_ClassInfo_CDC_Device_t CDC_Interface =
{
	.Config =
	{
		.ControlInterfaceNumber         = 0,

		.DataINEndpointNumber           = CDC_TX_EPNUM,
		.DataINEndpointSize             = CDC_TXRX_EPSIZE,
		.DataINEndpointDoubleBank       = false,

		.DataOUTEndpointNumber          = CDC_RX_EPNUM,
		.DataOUTEndpointSize            = CDC_TXRX_EPSIZE,
		.DataOUTEndpointDoubleBank      = false,

		.NotificationEndpointNumber     = CDC_NOTIFICATION_EPNUM,
		.NotificationEndpointSize       = CDC_NOTIFICATION_EPSIZE,
		.NotificationEndpointDoubleBank = false,
	}
	/*.State =
	{
		.ControlLineStates =
		{
			.HostToDevice				= USB_ControlRequest.wValue,
			.DeviceToHost				= USB_ControlRequest.wValue
		}
		.LineEncoding =
		{
			.BaudRateBPS				= 0,
			.CharFormat					= OneStopBit,
			.ParityType					= Parity_None,
			.DataBits					= 8
		}
	}*/
};

/*
 * Nur fr Debugzwecke -> da die seriellen Daten nur innerhalb des
 * AVR verarbeitet werden. D.h. der virtuelle COM-Port kann mit maximaler
 * USB-Geschwindigkeit betrieben werden. Auch die Flusskontrolle ist
 * irrelevant, da diese ber einen NAK Mechanismus in den Endpoint-
 * Transfer-Requests zum oder vom Host geregelt werden.
 */
CDC_Line_Coding_t MyLineEncoding =
{
	.BaudRateBPS = 128000,
	.CharFormat  = OneStopBit,
	.ParityType  = Parity_None,
	.DataBits    = 8
};


void HandleUSBDeviceState(USB_ClassInfo_CDC_Device_t* const Interface)
{
	if((USB_DeviceState != DEVICE_STATE_Configured)||(!(Interface->State.LineEncoding.BaudRateBPS)))
	{
	  return;
	}

	if(Endpoint_IsOUTReceived() && (!(Endpoint_BytesInEndpoint())))
	{
		Endpoint_ClearOUT();
	}

	CDC_Device_Flush(Interface);
}

/*
 ***************************************************************************
 *
 * Event-Handler fr Konfigurationsanfragen
 *
 ***************************************************************************
 */
void HandleControlRequest(USB_ClassInfo_CDC_Device_t* const CDCInterfaceInfo)
{
	if(!(Endpoint_IsSETUPReceived()))
	{
		return;
	}

	if(USB_ControlRequest.wIndex != CDCInterfaceInfo->Config.ControlInterfaceNumber)
	{
		return;
	}

	switch(USB_ControlRequest.bRequest)
	{
		case CDC_REQ_GetLineEncoding:
			if(USB_ControlRequest.bmRequestType == (REQDIR_DEVICETOHOST | REQTYPE_CLASS | REQREC_INTERFACE))
			{
				Endpoint_ClearSETUP();
				Endpoint_Write_Control_Stream_LE(&CDCInterfaceInfo->State.LineEncoding, sizeof(CDCInterfaceInfo->State.LineEncoding));
				Endpoint_ClearOUT();
			}
			break;

		case CDC_REQ_SetLineEncoding:
			if(USB_ControlRequest.bmRequestType == (REQDIR_HOSTTODEVICE | REQTYPE_CLASS | REQREC_INTERFACE))
			{
				Endpoint_ClearSETUP();
				Endpoint_Read_Control_Stream_LE(&CDCInterfaceInfo->State.LineEncoding, sizeof(CDCInterfaceInfo->State.LineEncoding));
				Endpoint_ClearIN();

				EVENT_CDC_Device_LineEncodingChanged(CDCInterfaceInfo);
			}
			break;

		case CDC_REQ_SetControlLineState:
			if(USB_ControlRequest.bmRequestType == (REQDIR_HOSTTODEVICE | REQTYPE_CLASS | REQREC_INTERFACE))
			{
				Endpoint_ClearSETUP();
				Endpoint_ClearStatusStage();

				CDCInterfaceInfo->State.ControlLineStates.HostToDevice = USB_ControlRequest.wValue;

				EVENT_CDC_Device_ControLineStateChanged(CDCInterfaceInfo);
			}
			break;

		case CDC_REQ_SendBreak:
			if(USB_ControlRequest.bmRequestType == (REQDIR_HOSTTODEVICE | REQTYPE_CLASS | REQREC_INTERFACE))
			{
				Endpoint_ClearSETUP();
				Endpoint_ClearStatusStage();

				EVENT_CDC_Device_BreakSent(CDCInterfaceInfo, (uint8_t)USB_ControlRequest.wValue);
			}
			break;
	}
}

/*
 ***************************************************************************
 *
 * USB Endpunkt Konfiguration
 *
 ***************************************************************************
 */
bool ConfigureEndpoints(USB_ClassInfo_CDC_Device_t* const CDCInterface)
{
	memset(&CDCInterface->State, 0x00, sizeof(CDCInterface->State));

	for(uint8_t EndpointNum = 1; EndpointNum < ENDPOINT_TOTAL_ENDPOINTS; EndpointNum++)
	{
		uint16_t Size;
		uint8_t  Type;
		uint8_t  Direction;
		bool     DoubleBanked;
		/*
		 * Enpoint #1 - Data-IN (bulk-transfer)
		 */
		if(EndpointNum == CDCInterface->Config.DataINEndpointNumber)
		{
			Size         = CDCInterface->Config.DataINEndpointSize;
			Direction    = ENDPOINT_DIR_IN;
			Type         = EP_TYPE_BULK;
			DoubleBanked = CDCInterface->Config.DataINEndpointDoubleBank;
		}
		/*
		 * Enpoint #1 - Data-OUT (bulk-transfer)
		 */
		else if(EndpointNum == CDCInterface->Config.DataOUTEndpointNumber)
		{
			Size         = CDCInterface->Config.DataOUTEndpointSize;
			Direction    = ENDPOINT_DIR_OUT;
			Type         = EP_TYPE_BULK;
			DoubleBanked = CDCInterface->Config.DataOUTEndpointDoubleBank;
		}
		/*
		 * Enpoint #1 - Notification/Config (interrupt-driven)
		 */
		else if(EndpointNum == CDCInterface->Config.NotificationEndpointNumber)
		{
			Size         = CDCInterface->Config.NotificationEndpointSize;
			Direction    = ENDPOINT_DIR_IN;
			Type         = EP_TYPE_INTERRUPT;
			DoubleBanked = CDCInterface->Config.NotificationEndpointDoubleBank;
		}
		else
		{
			continue;
		}

		if(!(Endpoint_ConfigureEndpoint(EndpointNum,Type,Direction,Size,DoubleBanked ? ENDPOINT_BANK_DOUBLE : ENDPOINT_BANK_SINGLE)))
		{
			return false;
		}
	}
	return true;
}


/*
 ***************************************************************************
 *
 * Setup-Funktionen zur Initialisierung der einzelnen Komponenten
 *
 ***************************************************************************
 */

void SetupUSB(void)
{
	// Watchdog deaktivieren
	MCUSR &= ~(1 << WDRF);
	wdt_disable();

	// Takt-Teiler auf 1 setzen (=deaktivieren)
	clock_prescale_set(clock_div_1);

	// LUFA USB initialisieren
	USB_Init();
}

void SetupSPI_Master(void)
{
	/*
	 * SPI Master Initialisierung:
	 * MOSI und SCK auf "output" schalten, alle anderen
	 * auf "input".
	 */
	DDRB = (1<<PB1)|(1<<PB2);

	/*
	 * SPE = Enable
	 * MSTR = Master(1)/Slave(0) Select
	 * 
	 * SPI2X   SPR1   SPR0    Frequency
	 *   0      0      0      f/4   = 4 Mhz
	 *   0      0      1      f/16  = 1 Mhz
	 *   0      1      0      f/64  = 250 Khz
	 *   0      1      1      f/128 = 125 Khz
	 *   1      0      0      f/2   = 8 Mhz
	 *   1      0      1      f/8   = 2 Mhz
	 *   1      1      0      f/32  = 500 Khz
	 *   1      1      1      f/64  = 250 Khz   
	 */

	//SPCR = ((1<<SPE)|(1<<MSTR)|(1<<SPR1));
	SPCR = ((1<<SPE)|(1<<MSTR)|(1<<SPR0)); //1 Mhz
	//SPCR = ((1<<SPE)|(1<<MSTR)|(1<<SPI2X)|(1<<SPR0)); //2 Mhz
	//SPCR = ((1<<SPE)|(1<<MSTR)); //4 Mhz
	
	/*
	 * Alle low-aktiven Slave-Select Leitungen an PORT D auf
	 * "output" schalten und auf HIGH ziehen.
	 */
	DDRD |= 0xFF;
    PORTD |= 0xFF;
}

void SetupSPI_Slave(void)
{
	/*
	 * SPI Slave Initialisierung:
	 * MISO auf "output" schalten, alle anderen
	 * auf "input".
	 */
	DDRB = (1<<PD3);
	
	/*
	 * SPE = Enable
	 * MSTR = Master/Slave Select
	 * Im Slave-Modus ist kein CLK Teiler ntig
	 */
	SPCR = (1<<SPE);
}

void SetupTimer(void)
{
	/*
	 * Timer Initialisierung:
	 *
	 * OCR0A = Output Compare Register (Max Value)
	 * WGM01 = CTC (Clear Timer on compare match)
	 * CS00 = CLK_I/O (no prescaling)
	 * CS01 = CLK_I/O / 8
	 * CS01|CS00 = CLK_I/O / 64
	 * CS02 = CLK_I/O / 256
	 * CS02|CS00 = CLK_I/O / 1024
	 * OCIE0A = Timer/Counter0 Output Compare Match A Interrupt Enable
	 */
	OCR0A  = 255;
	TCCR0A = (1<<WGM01);
	TCCR0B = (1<<CS02)|(1<<CS00);
	TIMSK0 = (1<<OCIE0A);
}

/*
 ***************************************************************************
 *
 * Interne Funktionen, die nur im Hauptprogramm verwendet werden
 *
 ***************************************************************************
 */

void SPI_Send(char cData)
{
	SPDR = cData;
	while(!(SPSR & (1<<SPIF)));
}

char SPI_Receive(void)
{
	while(!(SPSR & (1<<SPIF)));
	return SPDR;
}


void StreamToFramebuffer32(void)
{
	input_row       = (input_stream[0] & 0b11110000) >> 4;
	animation_speed = (input_stream[0] & 0b00001111);
	input_led       = (input_stream[1] & 0b11111111);

	/*
	 *		   Aufbau des 32 Bit Protokolls:
	 *    |---------|--------|---------|---------|
	 *       4    4     8      4    4    4    4
	 *    |____|____|________|____|____|____|____| 
	 *   Row-   Anim.    LED-   R    G    B  Update-
	 *  select  Speed   select               Flags
	 *
	 *
	 * Inhalt des Framebuffers: 0b0000RRRRGGGGBBBB
     *
	 */
    //framebuffer[input_row][input_led] = 0x0000;
	//framebuffer[input_row][input_led] = (input_stream[2] << 4) + (input_stream[3] >> 4);

	if(animation_speed > 0)
	{
		set_row = 0;
	}else{
		set_row = input_row;
	}

	framebuffer[set_row][input_led][R] = 0;
	framebuffer[set_row][input_led][G] = 0;
	framebuffer[set_row][input_led][B] = 0;

	framebuffer[set_row][input_led][R] = ((input_stream[2] & 0b11110000) >> 4);
	framebuffer[set_row][input_led][G] =  (input_stream[2] & 0b00001111);
	framebuffer[set_row][input_led][B] = ((input_stream[3] & 0b11110000) >> 4);
	
	control_byte = (input_stream[3] & 0b00001111);
}


void StreamToFramebuffer24(void)
{
	input_row       = ((input_stream[0] & 0b11110000) >> 4);
	animation_speed =  (input_stream[0] & 0b00001111);
	input_led       = ((input_stream[0] & 0b00000011) << 4)
	                + ((input_stream[1] & 0b11110000) >> 4) ;

	/*
	 *	   Alternatives 24 Bit Protokoll
	 *    |----------|--------|---------|
	 *      4    2   6      4    4    4  
	 *    |____|__|______|____|____|____| 
	 *    Row-  UP  LED-   R    G    B  
	 *   select    select             
	 *
	 *
	 * Inhalt des Framebuffers: 0b0000RRRRGGGGBBBB
	 *
	 */
    //framebuffer[input_row][input_led] = 0x0000;
	//framebuffer[input_row][input_led] = (input_stream[2] << 4) + (input_stream[3] >> 4);

	if(animation_speed > 0)
	{
		set_row = 0;
	}else{
		set_row = input_row;
	}

	framebuffer[set_row][input_led][R] = 0;
	framebuffer[set_row][input_led][G] = 0;
	framebuffer[set_row][input_led][B] = 0;

	framebuffer[set_row][input_led][R] = ((input_stream[2] & 0b11110000) >> 4);
	framebuffer[set_row][input_led][G] =  (input_stream[2] & 0b00001111);
	framebuffer[set_row][input_led][B] = ((input_stream[3] & 0b11110000) >> 4);

	control_byte = (input_stream[0] & 0b00000100);
}

void ClearFramebuffer(void)
{
	for(int row=0; row < 8; row++)
	{
		for(int led=0; led < 64; led++)
		{
			framebuffer[row][led][R] = 0;
			framebuffer[row][led][G] = 0;
			framebuffer[row][led][B] = 0;
		}
	}
}

void UpdateFrame(int rowcount)
{
	/*
	 * Die low-aktiven Slave-Select Ausgnge an PORTD
	 * sind One-Hot kodiert und mssen somit einzeln
	 * geschaltet werden.
	 */
	switch(rowcount)
	{
		case 0: setbit = 0x01;
				clearbit = 0xFE;
				break;

		case 1: setbit = 0x02;
				clearbit = 0xFD;
				break;

		case 2: setbit = 0x04;
				clearbit = 0xFB;
				break;
				
		case 3: setbit = 0x08;
				clearbit = 0xF7;
				break;
				
		case 4: setbit = 0x10;
				clearbit = 0xEF;
				break;
				
		case 5: setbit = 0x20;
				clearbit = 0xDF;
				break;
				
		case 6: setbit = 0x40;
				clearbit = 0xBF;
				break;
				
		case 7: setbit = 0x80;
				clearbit = 0x7F;
				break;
	}

	PORTD &= clearbit;

	for (int ledcount = 0; ledcount < 64; ++ledcount)
	{	
		SPI_Send((framebuffer[rowcount][ledcount][R] & 0b00001111));
		SPI_Send(((framebuffer[rowcount][ledcount][G] & 0b00001111) << 4)
				 +(framebuffer[rowcount][ledcount][B] & 0b00001111));
	}

	PORTD |= setbit;
}

void UpdateCube(void)
{
	for(int row=0; row < 8; row++)
	{
		UpdateFrame(row);
	}
}

void HandleInputStream(void)
{
	if(USBIsConnected)
	{
		StreamToFramebuffer32();
		//StreamToFramebuffer24();
	
		/*
		 * Update einer einzelnen Ebene (64 LEDs)
		 */
		if(control_byte < 0x08)
		{
			UpdateFrame(control_byte);
		}

		if(animation_speed > 0x00)
		{
			animation_enabled = true;
			//OCR0A  = animation_speed >> 4;
		}else{
			//OCR0A  = 255;
			animation_enabled = false;
		}

		/*
		 * Update des gesamten Cubes (512 LEDs)
		 */
		if(control_byte == 0x0F)
		{
			UpdateCube();
		}

		/*
		 * Lschen aller LEDs im Cube
		 */
		if(control_byte == 0x0C)
		{
			ClearFramebuffer();
			UpdateCube();
		}
	}
}

void DisplayNoConnectionWarning(void)
{
	ClearFramebuffer();
/*
	//Roter Pfeil (16 Bit Framebuffer)
	framebuffer[1][61] = 0x0F00;
	framebuffer[2][62] = 0x0F00;
	framebuffer[2][61] = 0x0F00;
	framebuffer[2][60] = 0x0F00;
	framebuffer[2][59] = 0x0F00;
	framebuffer[2][58] = 0x0F00;
	framebuffer[2][57] = 0x0F00;
	framebuffer[3][63] = 0x0F00;
	framebuffer[3][62] = 0x0F00;
	framebuffer[3][61] = 0x0F00;
	framebuffer[3][60] = 0x0F00;
	framebuffer[3][59] = 0x0F00;
	framebuffer[3][58] = 0x0F00;
	framebuffer[3][57] = 0x0F00;
	framebuffer[4][62] = 0x0F00;
	framebuffer[4][61] = 0x0F00;
	framebuffer[4][60] = 0x0F00;
	framebuffer[4][59] = 0x0F00;
	framebuffer[4][58] = 0x0F00;
	framebuffer[4][57] = 0x0F00;
	framebuffer[5][61] = 0x0F00;
*/

	//Roter Pfeil (8 Bit Framebuffer)
	framebuffer[1][61][R] = 0x0F;
	framebuffer[2][62][R] = 0x0F;
	framebuffer[2][61][R] = 0x0F;
	framebuffer[2][60][R] = 0x0F;
	framebuffer[2][59][R] = 0x0F;
	framebuffer[2][58][R] = 0x0F;
	framebuffer[2][57][R] = 0x0F;
	framebuffer[3][63][R] = 0x0F;
	framebuffer[3][62][R] = 0x0F;
	framebuffer[3][61][R] = 0x0F;
	framebuffer[3][60][R] = 0x0F;
	framebuffer[3][59][R] = 0x0F;
	framebuffer[3][58][R] = 0x0F;
	framebuffer[3][57][R] = 0x0F;
	framebuffer[4][62][R] = 0x0F;
	framebuffer[4][61][R] = 0x0F;
	framebuffer[4][60][R] = 0x0F;
	framebuffer[4][59][R] = 0x0F;
	framebuffer[4][58][R] = 0x0F;
	framebuffer[4][57][R] = 0x0F;
	framebuffer[5][61][R] = 0x0F;

	UpdateCube();
}

void DisplayConnectionReady(void)
{
	ClearFramebuffer();
	UpdateCube();
}

/*
 ***************************************************************************
 *
 * USB Event-Handler (siehe "LUFA_Events.lst")
 *
 ***************************************************************************
 */
void EVENT_USB_Device_ConfigurationChanged(void)
{
	if(ConfigureEndpoints(&CDC_Interface)) DisplayConnectionReady();
}

void EVENT_USB_Device_UnhandledControlRequest(void)
{
	HandleControlRequest(&CDC_Interface);
}

void EVENT_USB_Device_Connect(void)
{
	DisplayConnectionReady();
	USBIsConnected = true;
}

void EVENT_USB_Device_Disconnect(void)
{
	DisplayNoConnectionWarning();
	USBIsConnected = false;
}


/*
 ***************************************************************************
 *
 * Interrupt-Handler fr den Animations-Timer
 *
 ***************************************************************************
 */
ISR(TIMER0_COMPA_vect, ISR_BLOCK)
{
	for(int led=0; led < 64; led++)
	{
		framebuffer[row][led][R] = framebuffer[row-1][led][R];
		framebuffer[row][led][G] = framebuffer[row-1][led][G];
		framebuffer[row][led][B] = framebuffer[row-1][led][B];
	}
	UpdateFrame(row);

	if(++row == 8)
	{
		row = 1;
	}
}

/*
 ***************************************************************************
 *
 * Hauptprogramm
 *
 ***************************************************************************
 */
int main(void)
{
	SetupUSB();
	SetupSPI_Master();
	//SetupTimer();

	/*
	 * Character-Stream erstellen
	 */
	CDC_Device_CreateBlockingStream(&CDC_Interface, &USBSerialStream);
	
	//muss SEIn um globale Interrupts zu aktivieren
	sei();
	DisplayNoConnectionWarning();

	while(1)
	{
		/*
	     * 32 Bit Datenstrom byteweise einlesen:
		 * Funktion fgetc ist durch CreateBlockingStream
		 * berladen und blockiert bis neue Daten vorliegen.
	     */
		 if(USBIsConnected)
		 {
			input_stream[0] = fgetc(&USBSerialStream);
			input_stream[1] = fgetc(&USBSerialStream);
			input_stream[2] = fgetc(&USBSerialStream);
			input_stream[3] = fgetc(&USBSerialStream);

			/*
			 * Datenstrom entschlsseln und RGB Daten in Framebuffer ablegen
			 */
			HandleInputStream();
		}

		/*
		 * Gertestatus prfen und ggf. reinitialisieren
		 * Auerdem unbekannte Daten im Eingangspuffer
		 * verwerfen.
		 */
		HandleUSBDeviceState(&CDC_Interface);

		/*
		 * Enpunkte berprfen und ACK fr empfangene
		 * Daten senden.
		 */
		USB_USBTask();
	}
}
