/*
 * [S]erial [P]eripheral [I]nterface
 * RGB LED Cube Controller
 *
 *
 *
 * (c)2010
 * Oliver Kster
 * 1772570
 * Hochschule fr angewandte Wissenschaften Hamburg
 *
 *
 */

#include <avr/interrupt.h>

#define R 0
#define G 1
#define B 2

#define LE    7	
#define CLK   6
#define SDI   5
#define nOE   4
#define nB    3
#define nG    2
#define nR    1
#define hcLE  0

static volatile uint8_t framebuffer[3][64] = {{0},{0}};

volatile uint8_t SPICount = 0;
volatile uint8_t FBCount = 0;
volatile uint16_t ZeroCount = 0;
uint8_t spi_data[2] = {0};

/*
 ***************************************************************************
 *
 * Setup-Funktionen (werden nur bei der Initialisierung ein mal aufgerufen)
 *
 ***************************************************************************
 */

void SetupHardware(void)
{
	/*
	 * PORTD auf komplett "output" schalten
	 */
	DDRD |= 0xFF;

	/*
	 * LATCH ENABLE fr den einzelnen 74HC573
	 */
	PORTD |= (1<<hcLE);

	/*
	 * Low-aktives OUTPUT-ENABLE Bit 4 lschen
	 */
	PORTD &= ~(1<<nOE);
}


void SetupSPI_Slave(void)
{
	/*
	 * SPI Slave Initialisierung:
	 * MISO auf "output" schalten, alle anderen
	 * auf "input".
	 */
	DDRB = (1<<PB6);
	
	/*
	 * SPE = Enable
	 * MSTR = Master/Slave Select
	 * SPIE = SPI Interrupt Enable
	 * Im Slave-Modus ist kein CLK Teiler ntig
	 */
	SPCR |= (1<<SPE)|(1<<SPIE);
}

/*
 ***************************************************************************
 *
 * Interrupt-Handler fr das Empfangen von Nachrichten ber SPI
 *
 ***************************************************************************
 */
ISR(SPI_STC_vect)
{
	spi_data[SPICount] = SPDR;

	if(++SPICount == 2)
	{
		SPICount = 0;

		framebuffer[R][FBCount] = (spi_data[0] & 0b00001111);
		framebuffer[G][FBCount] = (spi_data[1] & 0b11110000) >> 4;
		framebuffer[B][FBCount] = (spi_data[1] & 0b00001111);

		if(++FBCount == 64)
		{
			FBCount = 0;
		}
	}
}

/*
 ***************************************************************************
 *
 * Hauptprogramm - PWM Generator (Originalfassung mit linearer Helligkeit)
 *
 ***************************************************************************
 */

int main(void)
{
	uint8_t led = 0;
	uint8_t color = 0;
	uint8_t pwm_value = 0;

	SetupHardware();
	SetupSPI_Slave();

	//muss SEIn um globale Interrupts zu aktivieren
	sei();


	while(1)
	{
		switch(color)
		{
			case 0:
				if(pwm_value < framebuffer[R][led])
				{
					PORTD |=  (1<<SDI)|(1<<CLK);
				}else{
					PORTD &= ~(1<<SDI);
					PORTD |= (1<<CLK);
				}
		
				if(++led == 64)
				{
					led = 0;
					PORTD |=  (1<<nG)|(1<<nB)|(1<<LE);
					PORTD &= ~(1<<nR);
			
					if(++pwm_value == 16)
					{
						pwm_value = 0;
						color = G;
					}
				}
			break;

			case 1:
				if(pwm_value < framebuffer[G][led])
				{
					PORTD |=  (1<<SDI)|(1<<CLK);
				}else{
					PORTD &= ~(1<<SDI);
					PORTD |= (1<<CLK);
				}
		
				if(++led == 64)
				{
					led = 0;
					PORTD |=  (1<<nR)|(1<<nB)|(1<<LE);
					PORTD &= ~(1<<nG);
			
					if(++pwm_value == 16)
					{
						pwm_value = 0;
						color = B;
					}
				}
			break;

			case 2:
				if(pwm_value < framebuffer[B][led])
				{
					PORTD |=  (1<<SDI)|(1<<CLK);
				}else{
					PORTD &= ~(1<<SDI);
					PORTD |= (1<<CLK);
				}
		
				if(++led == 64)
				{
					led = 0;
					PORTD |=  (1<<nR)|(1<<nG)|(1<<LE);
					PORTD &= ~(1<<nB);
			
					if(++pwm_value == 16)
					{
						pwm_value = 0;
						color = R;
					}
				}
			break;
		}
		PORTD &= ~((1<<CLK)|(1<<LE));
	}
}

/*
 ***************************************************************************
 *
 * Hauptprogramm - PWM Generator (Testversion mit Gammakorrektur)
 *
 ***************************************************************************
 */

int main_gamma_corrected(void)
{
	uint8_t led = 0;
	uint8_t color = 0;
	uint8_t pwm_value = 0;

	SetupHardware();
	SetupSPI_Slave();

	uint8_t PWMTable[16] = {0, 1, 2, 3, 5, 6, 7, 8,10,13,15,18,22,28,34,39};

	//muss SEIn um globale Interrupts zu aktivieren
	sei();


	while(1)
	{
		switch(color)
		{
			case 0:
				if(pwm_value < PWMTable[framebuffer[R][led]])
				{
					PORTD |=  (1<<SDI)|(1<<CLK);
				}else{
					PORTD &= ~(1<<SDI);
					PORTD |= (1<<CLK);
				}
		
				if(++led == 64)
				{
					led = 0;
					PORTD |=  (1<<nG)|(1<<nB)|(1<<LE);
					PORTD &= ~(1<<nR);
			
					if(++pwm_value == 40)
					{
						pwm_value = 0;
						color = G;
					}
				}
			break;

			case 1:
				if(pwm_value < PWMTable[framebuffer[G][led]])
				{
					PORTD |=  (1<<SDI)|(1<<CLK);
				}else{
					PORTD &= ~(1<<SDI);
					PORTD |= (1<<CLK);
				}
		
				if(++led == 64)
				{
					led = 0;
					PORTD |=  (1<<nR)|(1<<nB)|(1<<LE);
					PORTD &= ~(1<<nG);
			
					if(++pwm_value == 40)
					{
						pwm_value = 0;
						color = B;
					}
				}
			break;

			case 2:
				if(pwm_value < PWMTable[framebuffer[B][led]])
				{
					PORTD |=  (1<<SDI)|(1<<CLK);
				}else{
					PORTD &= ~(1<<SDI);
					PORTD |= (1<<CLK);
				}
		
				if(++led == 64)
				{
					led = 0;
					PORTD |=  (1<<nR)|(1<<nG)|(1<<LE);
					PORTD &= ~(1<<nB);
			
					if(++pwm_value == 40)
					{
						pwm_value = 0;
						color = R;
					}
				}
			break;
		}
		PORTD &= ~((1<<CLK)|(1<<LE));


		if ((PINB & (1<<PINB4)))
		{
			SPICount = 0;
			FBCount = 0;
		}
	}
}
